from typing import List
from typing import Optional

import pendulum

from tecton_proto.args.feature_view_pb2 import OfflineFeatureStoreConfig
from tecton_proto.args.feature_view_pb2 import OnlineStoreConfig
from tecton_proto.args.pipeline_pb2 import Pipeline
from tecton_proto.args.pipeline_pb2 import PipelineNode
from tecton_proto.data.feature_view_pb2 import FeatureView
from tecton_proto.data.feature_view_pb2 import TrailingTimeWindowAggregation
from tecton_proto.data.fv_materialization_pb2 import FvMaterialization
from tecton_proto.data.new_transformation_pb2 import NewTransformation as Transformation
from tecton_proto.data.virtual_data_source_pb2 import VirtualDataSource
from tecton_spark import time_utils
from tecton_spark.fco_container import FcoContainer
from tecton_spark.feature_view_utils import CONTINUOUS_MODE_BATCH_INTERVAL
from tecton_spark.feature_view_utils import get_input_feature_columns
from tecton_spark.feature_view_utils import validate_version
from tecton_spark.id_helper import IdHelper
from tecton_spark.logger import get_logger
from tecton_spark.online_serving_index import OnlineServingIndex
from tecton_spark.schema import Schema

logger = get_logger("FeatureDefinitionWrapper")


class FeatureDefinitionWrapper:
    def __init__(self, feature_view_proto: FeatureView, fco_container: FcoContainer):
        self.fv = feature_view_proto
        self.fco_container = fco_container

    @property
    def id(self) -> str:
        return IdHelper.to_string(self.fv.feature_view_id)

    @property
    def name(self) -> str:
        return self.fv.fco_metadata.name

    @property
    def is_temporal_aggregate(self) -> bool:
        return self.fv.HasField("temporal_aggregate")

    @property
    def is_continuous_temporal_aggregate(self) -> bool:
        return self.fv.temporal_aggregate.is_continuous

    @property
    def is_temporal(self) -> bool:
        return self.fv.HasField("temporal")

    @property
    def is_feature_table(self) -> bool:
        return self.fv.HasField("feature_table")

    @property
    def is_on_demand(self) -> bool:
        return self.fv.HasField("on_demand_feature_view")

    @property
    def get_feature_store_format_version(self) -> int:
        version = self.fv.feature_store_format_version
        validate_version(version)
        return version

    @property
    def time_key(self) -> bool:
        return self.fv.timestamp_key

    @property
    def join_keys(self) -> List[str]:
        return list(self.fv.join_keys)

    @property
    def online_serving_index(self) -> OnlineServingIndex:
        return OnlineServingIndex.from_proto(self.fv.online_serving_index)

    @property
    def wildcard_join_key(self) -> Optional[str]:
        """
        Returns a wildcard join key column name for the feature view if it exists;
        Otherwise returns None.
        """
        online_serving_index = self.online_serving_index
        wildcard_keys = [join_key for join_key in self.join_keys if join_key not in online_serving_index.join_keys]
        return wildcard_keys[0] if wildcard_keys else None

    @property
    def offline_store_config(self) -> OfflineFeatureStoreConfig:
        return self.fv.materialization_params.offline_store_config

    @property
    def online_store_params(self) -> OnlineStoreConfig:
        return self.fv.materialization_params.online_store_params

    @property
    def timestamp_key(self) -> Optional[str]:
        return self.fv.timestamp_key

    @property
    def materialization_enabled(self) -> bool:
        return self.fv.materialization_enabled

    @property
    def writes_to_offline_store(self) -> bool:
        if self.is_temporal_aggregate or self.is_temporal:
            return self.fv.materialization_params.writes_to_offline_store
        elif self.is_feature_table:
            return self.fv.feature_table.offline_enabled
        else:
            raise ValueError(f"Invalid invocation on unsupported FeatureView type")

    @property
    def writes_to_online_store(self) -> bool:
        if self.is_temporal_aggregate or self.is_temporal:
            return self.fv.materialization_params.writes_to_online_store
        elif self.is_feature_table:
            return self.fv.feature_table.online_enabled
        else:
            raise ValueError(f"Invalid invocation on unsupported FeatureView type")

    @property
    def view_schema(self) -> Schema:
        return Schema(self.fv.schemas.view_schema)

    @property
    def min_scheduling_interval(self) -> Optional[pendulum.Duration]:
        if self.is_feature_table:
            return None

        duration = None
        if self.is_temporal_aggregate:
            duration = self.fv.temporal_aggregate.slide_interval
        elif self.is_temporal:
            duration = self.fv.materialization_params.schedule_interval

        return time_utils.proto_to_duration(duration)

    @property
    def batch_materialization_schedule(self) -> pendulum.Duration:
        if self.is_temporal_aggregate and not self.fv.materialization_params.HasField("schedule_interval"):
            if self.is_continuous_temporal_aggregate:
                return time_utils.proto_to_duration(CONTINUOUS_MODE_BATCH_INTERVAL)
            else:
                return time_utils.proto_to_duration(self.fv.temporal_aggregate.slide_interval)
        else:
            return time_utils.proto_to_duration(self.fv.materialization_params.schedule_interval)

    @property
    def allowed_upstream_lateness(self) -> pendulum.Duration:
        return time_utils.proto_to_duration(self.fv.materialization_params.allowed_upstream_lateness)

    @property
    def start_timestamp(self) -> pendulum.datetime:
        return pendulum.instance(self.fv.materialization_params.start_timestamp.ToDatetime())

    @property
    def data_partitions_coalesce_override(self) -> int:
        return 10  # Value of DEFAULT_COALESCE_FOR_S3 as defined in materialization.py

    @property
    def data_source_ids(self) -> List[str]:
        return pipeline_to_ds_ids(self.fv.pipeline)

    @property
    def data_sources(self) -> List[VirtualDataSource]:
        ds_ids = pipeline_to_ds_ids(self.pipeline)
        return self.fco_container.get_by_ids(ds_ids)

    @property
    def get_tile_interval(self) -> pendulum.Duration:
        if self.is_temporal_aggregate:
            return time_utils.proto_to_duration(self.fv.temporal_aggregate.slide_interval)
        elif self.is_temporal:
            return time_utils.proto_to_duration(self.fv.materialization_params.schedule_interval)

        raise ValueError(f"Invalid invocation on unsupported FeatureView type")

    @property
    def get_aggregate_slide_interval_string(self) -> str:
        if self.is_temporal_aggregate:
            return self.fv.temporal_aggregate.slide_interval_string
        raise ValueError(f"Invalid invocation on unsupported FeatureView type")

    @property
    def fv_materialization(self) -> FvMaterialization:
        return self.fv.enrichments.fp_materialization

    @property
    def max_aggregation_window(self) -> Optional[int]:
        if self.is_temporal_aggregate:
            return max(
                [feature.window for feature in self.fv.temporal_aggregate.features],
                key=lambda window: window.ToSeconds(),
            )
        return None

    @property
    def transformations(self) -> List[Transformation]:
        transformation_ids = pipeline_to_transformation_ids(self.pipeline)
        return self.fco_container.get_by_ids(transformation_ids)

    @property
    def trailing_time_window_aggregation(self) -> Optional[TrailingTimeWindowAggregation]:
        if not self.is_temporal_aggregate:
            return None

        aggregation = TrailingTimeWindowAggregation()
        aggregation.time_key = self.timestamp_key
        slide_period_seconds = self.fv.temporal_aggregate.slide_interval.ToSeconds()
        aggregation.is_continuous = slide_period_seconds == 0
        aggregation.aggregation_slide_period.FromSeconds(slide_period_seconds)

        aggregation.features.extend(self.fv.temporal_aggregate.features)
        return aggregation

    @property
    def serving_ttl(self) -> Optional[pendulum.Duration]:
        if self.is_temporal:
            return time_utils.proto_to_duration(self.fv.temporal.serving_ttl)
        elif self.is_feature_table:
            return time_utils.proto_to_duration(self.fv.feature_table.serving_ttl)
        return None

    @property
    def features(self) -> List[str]:
        if self.is_temporal_aggregate and self.trailing_time_window_aggregation:
            return [
                f.output_feature_name
                for f in self.trailing_time_window_aggregation.features
                if f.output_feature_name != self.timestamp_key
            ]
        elif self.is_on_demand:
            return Schema(self.fv.schemas.view_schema).column_names()
        view_schema = Schema(self.fv.schemas.view_schema)
        return get_input_feature_columns(view_schema.to_proto(), self.join_keys, self.timestamp_key)

    @property
    def workspace(self) -> str:
        return self.fv.fco_metadata.workspace

    @property
    def pipeline(self) -> Pipeline:
        return self.fv.pipeline

    @property
    def feature_view_proto(self) -> FeatureView:
        return self.fv


def pipeline_to_ds_ids(pipeline: Pipeline) -> List[str]:
    ds_ids: List[str] = []

    def _recurse_pipeline_to_ds_ids(pipeline_node: PipelineNode, ds_ids_: List[str]):
        if pipeline_node.HasField("data_source_node"):
            _id = pipeline_node.data_source_node.virtual_data_source_id
            ds_id = IdHelper.to_string(_id)
            ds_ids_.append(ds_id)
        elif pipeline_node.HasField("transformation_node"):
            inputs = pipeline_node.transformation_node.inputs
            for input_ in inputs:
                _recurse_pipeline_to_ds_ids(input_.node, ds_ids_)

    _recurse_pipeline_to_ds_ids(pipeline.root, ds_ids)

    return ds_ids


def pipeline_to_transformation_ids(pipeline: Pipeline) -> List[str]:
    id_list: List[str] = []

    def _recurse_pipeline_to_transformation_ids(node: PipelineNode, id_list: List[str]):
        if node.HasField("transformation_node"):
            id_list.append(IdHelper.to_string(node.transformation_node.transformation_id))
            for input in node.transformation_node.inputs:
                _recurse_pipeline_to_transformation_ids(input.node, id_list)
        return id_list

    _recurse_pipeline_to_transformation_ids(pipeline.root, id_list)
    return id_list



from typing import Optional
from typing import Union

from tecton._internals import errors


class Constant:
    ALLOWED_TYPES = [str, int, float, bool, type(None)]

    def __init__(self, value: Optional[Union[str, int, float, bool]]):
        self.value = value
        self.value_type = type(value)

        if self.value_type not in self.ALLOWED_TYPES:
            raise errors.InvalidConstantType(value, self.ALLOWED_TYPES)

    def __repr__(self):
        return f"Constant(value={self.value}, type={self.value_type})"


def const(value: Optional[Union[str, int, float, bool]]) -> Constant:
    """
    Wraps a const and returns a ``Constant`` object that can be used inside pipeline functions.

    :param value: The constant value that needs to be wrapped and used in the pipeline function.
    :return: A Constant object.
    """
    return Constant(value)



import attr

from tecton_proto.args.feature_service_pb2 import LoggingConfigArgs


@attr.s(auto_attribs=True)
class LoggingConfig(object):
    """
    Configuration used to describe feature and request logging for Feature Services.

    :param sample_rate: (Optional, default 1.0) The rate of logging features. Must be between (0, 1]
    :param log_effective_times: (Optional) Whether to log the timestamps of the last update of the logged feature values

    An example of LoggingConfig declaration as part of FeatureService

    .. code-block:: Python

        from tecton import FeatureService, LoggingConfig

        # LoggingConfig is normaly used as a named argument parameter to a FeatureService instance definition.
        my_feature_service = FeatureService(
            name="An example of Feature Service"
            # Other named arguments
            ...
            # A LoggingConfig instance
            logging=LoggingConfig(
                    sample_rate=0.5,
                    log_effective_time=False,
            )
            ...
        )
    """

    sample_rate: float = attr.ib(default=1.0, validator=attr.validators.instance_of((float, int)))
    log_effective_times: bool = attr.ib(default=False, validator=attr.validators.instance_of(bool))

    def _to_proto(self) -> LoggingConfigArgs:
        return LoggingConfigArgs(sample_rate=self.sample_rate, log_effective_times=self.log_effective_times)

    @classmethod
    def _from_proto(cls, logging_config_proto: LoggingConfigArgs) -> "LoggingConfig":
        return LoggingConfig(logging_config_proto.sample_rate, logging_config_proto.log_effective_times)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/common/spark_schema.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='tecton_proto/common/spark_schema.proto',
  package='tecton_proto.common',
  syntax='proto2',
  serialized_options=b'\n\021com.tecton.commonP\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n&tecton_proto/common/spark_schema.proto\x12\x13tecton_proto.common\"K\n\nSparkField\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12)\n\x10structfield_json\x18\x02 \x01(\tR\x0fstructfieldJson\"F\n\x0bSparkSchema\x12\x37\n\x06\x66ields\x18\x01 \x03(\x0b\x32\x1f.tecton_proto.common.SparkFieldR\x06\x66ieldsB\x15\n\x11\x63om.tecton.commonP\x01'
)




_SPARKFIELD = _descriptor.Descriptor(
  name='SparkField',
  full_name='tecton_proto.common.SparkField',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='tecton_proto.common.SparkField.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='name', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='structfield_json', full_name='tecton_proto.common.SparkField.structfield_json', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='structfieldJson', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=63,
  serialized_end=138,
)


_SPARKSCHEMA = _descriptor.Descriptor(
  name='SparkSchema',
  full_name='tecton_proto.common.SparkSchema',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='fields', full_name='tecton_proto.common.SparkSchema.fields', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='fields', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=140,
  serialized_end=210,
)

_SPARKSCHEMA.fields_by_name['fields'].message_type = _SPARKFIELD
DESCRIPTOR.message_types_by_name['SparkField'] = _SPARKFIELD
DESCRIPTOR.message_types_by_name['SparkSchema'] = _SPARKSCHEMA
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

SparkField = _reflection.GeneratedProtocolMessageType('SparkField', (_message.Message,), {
  'DESCRIPTOR' : _SPARKFIELD,
  '__module__' : 'tecton_proto.common.spark_schema_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.common.SparkField)
  })
_sym_db.RegisterMessage(SparkField)

SparkSchema = _reflection.GeneratedProtocolMessageType('SparkSchema', (_message.Message,), {
  'DESCRIPTOR' : _SPARKSCHEMA,
  '__module__' : 'tecton_proto.common.spark_schema_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.common.SparkSchema)
  })
_sym_db.RegisterMessage(SparkSchema)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/common/schema.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.common import column_type_pb2 as tecton__proto_dot_common_dot_column__type__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='tecton_proto/common/schema.proto',
  package='tecton_proto.common',
  syntax='proto2',
  serialized_options=b'\n\021com.tecton.commonP\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n tecton_proto/common/schema.proto\x12\x13tecton_proto.common\x1a%tecton_proto/common/column_type.proto\"\x93\x01\n\x06\x43olumn\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12$\n\x0eraw_spark_type\x18\x02 \x01(\tR\x0crawSparkType\x12O\n\x13\x66\x65\x61ture_server_type\x18\x03 \x01(\x0e\x32\x1f.tecton_proto.common.ColumnTypeR\x11\x66\x65\x61tureServerType\"E\n\x06Schema\x12\x35\n\x07\x63olumns\x18\x02 \x03(\x0b\x32\x1b.tecton_proto.common.ColumnR\x07\x63olumnsJ\x04\x08\x01\x10\x02\x42\x15\n\x11\x63om.tecton.commonP\x01'
  ,
  dependencies=[tecton__proto_dot_common_dot_column__type__pb2.DESCRIPTOR,])




_COLUMN = _descriptor.Descriptor(
  name='Column',
  full_name='tecton_proto.common.Column',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='tecton_proto.common.Column.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='name', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='raw_spark_type', full_name='tecton_proto.common.Column.raw_spark_type', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='rawSparkType', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_server_type', full_name='tecton_proto.common.Column.feature_server_type', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='featureServerType', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=97,
  serialized_end=244,
)


_SCHEMA = _descriptor.Descriptor(
  name='Schema',
  full_name='tecton_proto.common.Schema',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='columns', full_name='tecton_proto.common.Schema.columns', index=0,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='columns', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=246,
  serialized_end=315,
)

_COLUMN.fields_by_name['feature_server_type'].enum_type = tecton__proto_dot_common_dot_column__type__pb2._COLUMNTYPE
_SCHEMA.fields_by_name['columns'].message_type = _COLUMN
DESCRIPTOR.message_types_by_name['Column'] = _COLUMN
DESCRIPTOR.message_types_by_name['Schema'] = _SCHEMA
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Column = _reflection.GeneratedProtocolMessageType('Column', (_message.Message,), {
  'DESCRIPTOR' : _COLUMN,
  '__module__' : 'tecton_proto.common.schema_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.common.Column)
  })
_sym_db.RegisterMessage(Column)

Schema = _reflection.GeneratedProtocolMessageType('Schema', (_message.Message,), {
  'DESCRIPTOR' : _SCHEMA,
  '__module__' : 'tecton_proto.common.schema_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.common.Schema)
  })
_sym_db.RegisterMessage(Schema)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/user.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='tecton_proto/data/user.proto',
  package='tecton_proto.data',
  syntax='proto2',
  serialized_options=b'\n\017com.tecton.dataP\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1ctecton_proto/data/user.proto\x12\x11tecton_proto.data\x1a\x1fgoogle/protobuf/timestamp.proto\"\xae\x02\n\x04User\x12\x17\n\x07okta_id\x18\x01 \x01(\tR\x06oktaId\x12\x1d\n\nfirst_name\x18\x02 \x01(\tR\tfirstName\x12\x1b\n\tlast_name\x18\x03 \x01(\tR\x08lastName\x12\x1f\n\x0blogin_email\x18\x04 \x01(\tR\nloginEmail\x12\x1f\n\x0bokta_status\x18\x05 \x01(\tR\noktaStatus\x12\x39\n\ncreated_at\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tcreatedAt\x12\x19\n\x08is_admin\x18\x07 \x01(\x08R\x07isAdmin\x12\x39\n\nlast_login\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tlastLoginB\x13\n\x0f\x63om.tecton.dataP\x01'
  ,
  dependencies=[google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,])




_USER = _descriptor.Descriptor(
  name='User',
  full_name='tecton_proto.data.User',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='okta_id', full_name='tecton_proto.data.User.okta_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='oktaId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='first_name', full_name='tecton_proto.data.User.first_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='firstName', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_name', full_name='tecton_proto.data.User.last_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='lastName', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='login_email', full_name='tecton_proto.data.User.login_email', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='loginEmail', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='okta_status', full_name='tecton_proto.data.User.okta_status', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='oktaStatus', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='created_at', full_name='tecton_proto.data.User.created_at', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='createdAt', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='is_admin', full_name='tecton_proto.data.User.is_admin', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='isAdmin', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_login', full_name='tecton_proto.data.User.last_login', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='lastLogin', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=85,
  serialized_end=387,
)

_USER.fields_by_name['created_at'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_USER.fields_by_name['last_login'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
DESCRIPTOR.message_types_by_name['User'] = _USER
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

User = _reflection.GeneratedProtocolMessageType('User', (_message.Message,), {
  'DESCRIPTOR' : _USER,
  '__module__' : 'tecton_proto.data.user_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.data.User)
  })
_sym_db.RegisterMessage(User)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



import functools
from dataclasses import dataclass
from textwrap import dedent

from tecton._internals.fco import Fco
from tecton.transformations.transformation import Transformation
from tecton_proto.args.pipeline_pb2 import PipelineNode
from tecton_spark.materialization_context import materialization_context


@dataclass
class BuiltinTransformation(Transformation):
    """
    Note that you SHOULD NOT change the code of any builtin transformations since it will cause
    errors to be thrown in MDS validation since we cannot allow destructive recreations.
    """

    docstring: str
    call_count: int = 0

    def _docstring(self):
        return self.docstring

    def _is_builtin(self):
        return True

    def __call__(self, *args, **kwargs) -> PipelineNode:
        # register the first time it's called
        if self.call_count == 0:
            Fco._register(self)
        self.call_count += 1
        return super().__call__(*args, **kwargs)

    def __hash__(self):
        return hash(self.name)


def builtin_transformation(mode: str, description: str, docstring: str):
    def decorator(user_function):
        transform_name = user_function.__name__
        transform = BuiltinTransformation(
            transform_name,
            mode,
            user_function,
            description,
            owner="tecton",
            family="tecton_builtins",
            tags={},
            docstring=dedent(docstring),
        )
        functools.update_wrapper(wrapper=transform, wrapped=user_function)
        return transform

    return decorator


@builtin_transformation(
    mode="pyspark",
    description="A transformation that adds a column to make sliding window aggregations easier.",
    docstring="""
        :param df: Spark DataFrame
        :param timestamp_key: The name of the timestamp columns for the event times in `df`
        :param window_size: How long each sliding window is, as a string in the format "[QUANTITY] [UNIT]".
            Ex: "2 days". See https://pypi.org/project/pytimeparse/ for more details.
        :param slide_interval: [optional] How often window is produced, as a string in the format "[QUANTITY] [UNIT]".
            Ex: "2 days". See https://pypi.org/project/pytimeparse/ for more details.
            Note this must be less than or equal to window_size, and window_size must be a multiple of slide_interval.
            If not provided, this defaults to the batch schedule of the FeatureView.
        :param window_column_name: [optional] The output column name for the timestamp of the end of each window
        :return: An exploded Spark DataFrame with an added column according to window_column_name.

        Ex:
            tecton_sliding_window(
                [
                    (user_id=1, timestamp = '2021-01-10 10:14:14'),
                    (user_id=2, timestamp = '2021-01-11 23:10:10'),
                    (user_id=3, timestamp = '2021-01-12 01:01:01')
                ],
                window_size = '2 days')
            with context(
                feature_start_time='2021-01-10 00:00:00',
                feature_end_time='2021-01-12 00:00:00',
                batch_schedule='1 day'
            )
            =>
            [
                (user_id=1, timestamp='2021-01-10 10:14:14', window_end='2021-01-10 23:59:59.999999'),
                (user_id=1, timestamp='2021-01-10 10:14:14', window_end='2021-01-11 23:59:59.999999'),
                (user_id=2, timestamp='2021-01-11 23:10:10', window_end='2021-01-11 23:59:59.999999')
            ]

            Note that each input row can produce from 0 to (window_size / slide_interval) output rows, depending on how many
            windows it is present in.

            To do an aggregation on these rows, you could write a query like below, which computes a count for each window.
            ```
            SELECT
                user_id,
                COUNT(1) as occurrences,
                window_end
            FROM {df}
            GROUP BY
                user_id,
                window_end
            ```
    """,
)
def tecton_sliding_window(
    df,
    timestamp_key,
    window_size,
    slide_interval=None,
    window_column_name="window_end",
    context=materialization_context(),
):
    from pyspark.sql import functions as F
    from tecton_spark.udfs import tecton_sliding_window_udf, _validate_sliding_window_duration

    slide_interval = slide_interval or f"{context.batch_schedule.total_seconds()} seconds"
    _validate_sliding_window_duration(window_size, slide_interval)

    return df.withColumn(
        window_column_name,
        F.explode(
            tecton_sliding_window_udf(
                F.col(timestamp_key),
                F.lit(window_size),
                F.lit(slide_interval),
                F.lit(context.feature_start_time),
                F.lit(context.feature_end_time),
            )
        ),
    )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/args/repo_metadata.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.common import id_pb2 as tecton__proto_dot_common_dot_id__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='tecton_proto/args/repo_metadata.proto',
  package='tecton_proto.args',
  syntax='proto2',
  serialized_options=b'\n\017com.tecton.argsP\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n%tecton_proto/args/repo_metadata.proto\x12\x11tecton_proto.args\x1a\x1ctecton_proto/common/id.proto\"\xa6\x01\n\nSourceInfo\x12.\n\x06\x66\x63o_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x05\x66\x63oId\x12\x14\n\x05scope\x18\x03 \x01(\tR\x05scope\x12#\n\rsource_lineno\x18\x04 \x01(\tR\x0csourceLineno\x12\'\n\x0fsource_filename\x18\x05 \x01(\tR\x0esourceFilenameJ\x04\x08\x02\x10\x03\"W\n\x15\x46\x65\x61tureRepoSourceInfo\x12>\n\x0bsource_info\x18\x01 \x03(\x0b\x32\x1d.tecton_proto.args.SourceInfoR\nsourceInfoB\x13\n\x0f\x63om.tecton.argsP\x01'
  ,
  dependencies=[tecton__proto_dot_common_dot_id__pb2.DESCRIPTOR,])




_SOURCEINFO = _descriptor.Descriptor(
  name='SourceInfo',
  full_name='tecton_proto.args.SourceInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='fco_id', full_name='tecton_proto.args.SourceInfo.fco_id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='fcoId', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='scope', full_name='tecton_proto.args.SourceInfo.scope', index=1,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='scope', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='source_lineno', full_name='tecton_proto.args.SourceInfo.source_lineno', index=2,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='sourceLineno', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='source_filename', full_name='tecton_proto.args.SourceInfo.source_filename', index=3,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='sourceFilename', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=91,
  serialized_end=257,
)


_FEATUREREPOSOURCEINFO = _descriptor.Descriptor(
  name='FeatureRepoSourceInfo',
  full_name='tecton_proto.args.FeatureRepoSourceInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='source_info', full_name='tecton_proto.args.FeatureRepoSourceInfo.source_info', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='sourceInfo', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=259,
  serialized_end=346,
)

_SOURCEINFO.fields_by_name['fco_id'].message_type = tecton__proto_dot_common_dot_id__pb2._ID
_FEATUREREPOSOURCEINFO.fields_by_name['source_info'].message_type = _SOURCEINFO
DESCRIPTOR.message_types_by_name['SourceInfo'] = _SOURCEINFO
DESCRIPTOR.message_types_by_name['FeatureRepoSourceInfo'] = _FEATUREREPOSOURCEINFO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

SourceInfo = _reflection.GeneratedProtocolMessageType('SourceInfo', (_message.Message,), {
  'DESCRIPTOR' : _SOURCEINFO,
  '__module__' : 'tecton_proto.args.repo_metadata_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.args.SourceInfo)
  })
_sym_db.RegisterMessage(SourceInfo)

FeatureRepoSourceInfo = _reflection.GeneratedProtocolMessageType('FeatureRepoSourceInfo', (_message.Message,), {
  'DESCRIPTOR' : _FEATUREREPOSOURCEINFO,
  '__module__' : 'tecton_proto.args.repo_metadata_pb2'
  # @@protoc_insertion_point(class_scope:tecton_proto.args.FeatureRepoSourceInfo)
  })
_sym_db.RegisterMessage(FeatureRepoSourceInfo)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



from typing import Dict

from tecton._internals import utils
from tecton._internals.sdk_decorators import sdk_public_method


class Fco:
    """
    The class primarily intended for interactively initialized objects (with the exception of the name method)
    """

    @classmethod
    def _fco_type_name_singular_snake_case(cls) -> str:
        raise NotImplementedError

    @classmethod
    def _fco_type_name_plural_snake_case(cls) -> str:
        raise NotImplementedError

    @classmethod
    def _fco_type_name_singular_capitalized(cls) -> str:
        return utils.snake_to_capitalized(cls._fco_type_name_singular_snake_case())

    @classmethod
    def _fco_type_name_plural_capitalized(cls) -> str:
        return utils.snake_to_capitalized(cls._fco_type_name_plural_snake_case())

    @property
    def _fco_metadata(self):
        raise Exception("unimplemented _fco_metadata property")

    @property  # type: ignore
    @sdk_public_method
    def name(self) -> str:
        """
        The name of this Tecton Object.
        """
        # This method supports declaratively initialized objects as it is convenient to use
        # in declarative repo for linking FCOs by name
        return self._args.info.name if hasattr(self, "_args") else self._fco_metadata.name  # type: ignore

    @property  # type: ignore
    @sdk_public_method
    def description(self) -> str:
        """
        The description of this Tecton Object, set by user.
        """
        return self._fco_metadata.description

    @property  # type: ignore
    @sdk_public_method
    def family(self) -> str:
        """
        The family of this Tecton Object, used to group Objects.
        """
        return self._fco_metadata.family

    @property  # type: ignore
    @sdk_public_method
    def tags(self) -> Dict[str, str]:
        """
        Tags associated with this Tecton Object (key-value pairs of arbitrary metadata set by user.)
        """
        return dict(self._fco_metadata.tags)

    @property  # type: ignore
    @sdk_public_method
    def owner(self) -> str:
        """
        The owner of this Tecton Object (typically the email of the primary maintainer.)
        """
        return self._fco_metadata.owner

    @property  # type: ignore
    @sdk_public_method
    def created_at(self) -> str:
        """
        Returns the creation date of this Tecton Object.
        """
        return self._fco_metadata.created_at.ToDatetime().strftime("%Y-%m-%d %H:%M:%S")

    @property  # type: ignore
    @sdk_public_method
    def defined_in(self) -> str:
        """
        Returns filename where this Tecton Object has been declared.
        """
        return self._fco_metadata.source_filename

    @property
    def workspace(self) -> str:
        """
        Returns the workspace this Tecton Object was created in.
        """
        return self._fco_metadata.workspace



from dataclasses import dataclass
from typing import List

import py4j.protocol
from pyspark.sql import DataFrame

from tecton_spark.logger import get_logger


logger = get_logger("SparkHelper")


def _get_table_names_recursively(node):
    """Return the list of table names extracted from the Spark's logical plan (tree rooted at node)."""
    try:
        if node.nodeName() == "UnresolvedRelation":
            return [node.tableName()]
        else:
            children = node.children()
            child_nodes = []
            # node and children are Java objects and working with them in Python is really weird.
            # This is the only hacky way I found to traverse the tree.
            for idx in range(children.size()):
                child_nodes.append(children.slice(idx, idx + 1).last())
            tables = []
            for child in child_nodes:
                tables.extend(_get_table_names_recursively(child))
            return tables
    except py4j.protocol.Py4JError as e:
        # Log the error, since this shouldn't happen. However, if for some weird queries the parsing fails,
        # we should try our best to return as many parsed tables as possible.
        logger.error(f"Failed to get table names recursively for node:\n{node}With exception:\n{e}")
        return []


@dataclass
class QueryPlanInfo:
    has_joins: bool
    has_aggregations: bool
    table_names: List[str]


def _has_node(node, name: str):
    if node.nodeName() == name:
        return True
    else:
        children = node.children()
        for idx in range(children.size()):
            if _has_node(children.slice(idx, idx + 1).last(), name):
                return True
    return False


def get_query_plan_info_for_df(df: DataFrame) -> QueryPlanInfo:
    plan = df._jdf.queryExecution().logical()
    table_names = _get_table_names_recursively(plan)
    return QueryPlanInfo(
        table_names=table_names, has_joins=_has_node(plan, "Join"), has_aggregations=_has_node(plan, "Aggregate")
    )



class TectonAPIValidationError(ValueError):
    """
    Exception that indicates a problem in validating user inputs against
    the data in the system. Typically recoverable by the user.
    """


class TectonAPIInaccessibleError(Exception):
    """
    Exception that indicates a problem connecting to Tecton cluster.
    """


class FailedPreconditionError(Exception):
    """
    Exception that indicates some prequisite has not been met (e.g the CLI/SDK needs to be updated).
    """
